{$APPTYPE CONSOLE}
{$r+}

program flexion;

  {  Calcul statique d'une poutre droite sollicite en flexion simple
     par la mthode des lments finis

     Programme crit en PASCAL - Application console
     Compilateur DELPHI
     mai 2009

     Yves DEBARD
     Dpartement Gnie Mcanique et Productique
     Institut Universitaire de Technologie du MANS

  }

uses SysUtils,Windows;

const

  Max_noeuds = 50 ;
  Max_elements = Max_noeuds-1;
  Max_liaisons = 20 ;
  Max_charges =20 ;
  NDDLE = 4 ;          { nombre de degrs de libert d'un lment }
  Max_ddl = 2*Max_noeuds ;

  ID_DY_NUL = 1 ;
  ID_ROTZ_NUL = 2 ;
  ID_ENCASTREMENT = 3 ;

  id_force_nodale = 1 ;
  id_force_repartie = 2 ;

  Max_points = 10 ;

type

  String80 = string[80];

  TVecteur_elementaire = array [1..NDDLE] of Double;
  TMatrice_elementaire = array [1..NDDLE,1..NDDLE] of Double;
  TLocalisation_ddl_elementaires = array [1..NDDLE] of Integer;
  TVecteur_global = array [1..Max_ddl] of Double;
  TMatrice_globale = array [1..Max_ddl,1..Max_ddl] of Double;

  TGraphe_elementaire = record
      NPoints:Integer;
      x,depl_y,rot_z,ty,mfz:array [0..Max_points] of Double;
    end;
  TElement = class
      Origine,Extremite:Integer;
      Young,Iz:Double;
      Fe:TVecteur_elementaire;
      constructor Create;
      function Longueur:Double;
      function Rigidite(var ke:TMatrice_elementaire):Boolean;
      procedure Graphe(var g:TGraphe_elementaire);
    end;
  TElements = class
      Nombre:Integer;
      Table:array [1..Max_elements] of TElement;
      constructor Create;
      destructor Destroy;override;
      function Index(e:TElement):Integer;
    end;
  TLiaison = class
      Type_liaison,Noeud:Integer;
      Ry,Mz:Double;
      constructor Create;
    end;
  TLiaisons = class
      Nombre:Integer;
      Table:array [1..Max_liaisons] of TLiaison;
      constructor Create;
      destructor Destroy;override;
    end;
  TCharge = class
      Type_charge,Entite:Integer;
      Fy,Mz:Double;
      constructor Create;
    end;
  TCharges = class
      Nombre:Integer;
      Table:array [1..Max_charges] of TCharge;
      constructor Create;
      destructor Destroy;override;
    end;

  TStructure = class
      Nom_du_projet:String80;
      NNoeuds,NDdl:Integer;
      Noeuds:array [1..Max_noeuds] of Double;
      Elements:TElements;
      Liaisons:TLiaisons;
      Charges:TCharges;
      Matkg:TMatrice_globale;
      Vectg:TVecteur_global;
      fres:Text;
      Fichier_resultat:Boolean;
      constructor Create;
      destructor Destroy;override;
      function Lire_les_donnees:Boolean;
      procedure Localisation_ddl_noeud(i:Integer;
                                          var ddl_uy,ddl_rotz:Integer);
      procedure Localisation_ddl_element(i:Integer;
                                   var locddl:TLocalisation_ddl_elementaires);
      function Assembler_la_matrice_de_rigidite:Boolean;
      procedure Assembler_le_vecteur_force;
      procedure Limites;
      function Calculer_les_deplacements:Boolean;
      procedure Efforts_elementaires;
      procedure Actions_de_liaison;
      procedure Editions;
      procedure Calculer;
    end;

var Structure:TStructure;

//==============================================================================

constructor TElement.Create;
  begin
    inherited Create;
  end;

function TElement.Longueur:Double;
  begin
    result:=Structure.Noeuds[Extremite]-Structure.Noeuds[Origine];
  end;

function TElement.Rigidite(var ke:TMatrice_elementaire):Boolean;
  var long,long2,x:Double;
      j,k:Integer;
  begin
    long:=Longueur;
    if long<1e-6 then begin result:=False;Exit;end;
    long2:=Sqr(long);
    x:=Young*Iz/(long2*long);
    ke[1,1]:=x*12;ke[1,2]:=x*6*long;ke[1,3]:=-X*12;ke[1,4]:=x*6*long;
    ke[2,2]:=x*4*long2;ke[2,3]:=-x*6*long;ke[2,4]:=x*2*long2;
    ke[3,3]:=x*12;ke[3,4]:=-x*6*long;
    ke[4,4]:=x*4*long2;
    for j:=2 to 4 do for k:=1 to j-1 do ke[j,k]:=ke[k,j];
    result:=True;
  end;

procedure TElement.Graphe(var g:TGraphe_elementaire);

  var fy,c,ksi,long,depl_o,rot_o,depl_e,rot_e:Double;
      k:Integer;
      charge:TCharge;

  begin

    FillChar(g,Sizeof(g),#0);
    g.NPoints:=Max_points;

    depl_o:=Structure.Vectg[2*Origine-1];
    rot_o:=Structure.Vectg[2*Origine];
    depl_e:=Structure.Vectg[2*Extremite-1];
    rot_e:=Structure.Vectg[2*Extremite];

    fy:=0;
    for k:=1 to Structure.Charges.Nombre do
      begin
        charge:=Structure.Charges.Table[k];
        if (charge.Type_charge=id_force_repartie) and (charge.Entite=Structure.Elements.Index(Self))
          then fy:=fy+charge.Fy;
      end;

    long:=Longueur;

    for k:=0 to g.NPoints do
      begin
        ksi:=k/g.NPoints;
        g.x[k]:=Structure.Noeuds[Origine]+ksi*long;
        g.depl_y[k]:=depl_o*(1-3*Sqr(ksi)+2*Sqr(ksi)*ksi)
                     +rot_o*(ksi-2*Sqr(ksi)+Sqr(ksi)*ksi)*long
                     +depl_e*(3*Sqr(ksi)-2*Sqr(ksi)*ksi)
                     +rot_e*(-Sqr(ksi)+Sqr(ksi)*ksi)*long;
        g.rot_z[k]:=depl_o*6*(-ksi+Sqr(ksi))/long
                    +rot_o*(1-4*ksi+3*Sqr(ksi))
                    +depl_e*6*(ksi-Sqr(ksi))/long
                    +rot_e*(-2*ksi+3*Sqr(ksi));
        g.ty[k]:=Fe[3];
        g.mfz[k]:=Fe[4]+long*Fe[3]*(1-ksi);
        if Abs(fy)>1e-6
          then
            begin
              c:=fy*Sqr(long)*long/(12*Young*Iz);
              g.depl_y[k]:=g.depl_y[k]+0.5*c*long*Sqr(ksi)*Sqr(1-ksi);
              g.rot_z[k]:=g.rot_z[k]+c*ksi*(1-ksi)*(1-2*ksi);
              g.ty[k]:=g.ty[k]+fy*long*(1-ksi);
              g.mfz[k]:=g.mfz[k]+0.5*fy*Sqr(long)*Sqr(1-ksi);
            end;
      end;

  end;


constructor TElements.Create;
  begin
    inherited Create;
  end;

destructor TElements.Destroy;
  var i:Integer;
  begin
    for i:=1 to Nombre do Table[i].Free;
    inherited Destroy;
  end;

function TElements.Index(e:TElement):Integer;
  var i:Integer;
  begin
    for i:=1 to Nombre do
      if Table[i]=e then begin result:=i;Exit;end;
    result:=0;
  end;

constructor TLiaison.Create;
  begin
    inherited Create;
  end;

constructor TLiaisons.Create;
  begin
    inherited Create;
  end;

destructor TLiaisons.Destroy;
  var i:Integer;
  begin
    for i:=1 to Nombre do Table[i].Free;
    inherited Destroy;
  end;

constructor TCharge.Create;
  begin
    inherited Create;
  end;

constructor TCharges.Create;
  begin
    inherited Create;
  end;

destructor TCharges.Destroy;
  var i:Integer;
  begin
    for i:=1 to Nombre do Table[i].Free;
    inherited Destroy;
  end;

//==============================================================================

constructor TStructure.Create;
  begin
    inherited Create;
    Elements:=TElements.Create;
    Liaisons:=TLiaisons.Create;
    Charges:=TCharges.Create;
    Fichier_resultat:=False;
  end;

destructor TStructure.Destroy;
  var s:String;
  begin
    Elements.Free;
    Liaisons.Free;
    Charges.Free;
    if Fichier_resultat
      then
        begin
          close(fres);
          s:='notepad '+nom_du_projet+'r.txt';
          Winexec(PChar(s),SW_NORMAL);
        end;
    inherited Destroy;
  end;


function TStructure.Lire_les_donnees:Boolean;

// les donnes sont lues dans un fichier texte dont l'extension est .txt

  var i:Integer;
      nom_fichier,ligne:String80;
      f:text;

  function Mot_cle(mc,s:String80):Boolean;
    var k:Integer;
        s1,s2:String80;
    begin
      s1:='';for k:=1 to Length(mc) do if mc[k]<>' ' then s1:=s1+Upcase(mc[k]);
      s2:='';for k:=1 to Length(s) do if s[k]<>' ' then s2:=s2+Upcase(s[k]);
      s2:=copy(s2,1,Length(s1));
      result:=s1=s2;
    end;

  begin

    result:=True;

    write('Nom du fichier ( sans extension ) ? ');
    readln(Nom_du_projet);
    if Nom_du_projet=''
      then begin result:=False;Exit;end;
    nom_fichier:=Nom_du_projet+'.txt';

    assign(f,nom_fichier);{$i-}reset(f);{$i+}
    if Ioresult<>0 then begin result:=False;Exit;end;

    assign(fres,nom_du_projet+'r.txt');{$i-}rewrite(fres);{$i+}
    if Ioresult<>0 then begin close(f);result:=False;Exit;end;
    Fichier_resultat:=True;


    writeln(fres,'Lecture du fichier : ',nom_fichier);

    repeat
      if Eof(f) then begin result:=False;Close(f);Exit;end;
      readln(f,ligne);
    until Mot_cle('$deb',ligne);

    repeat

      if Eof(f) then begin result:=False;Close(f);Exit;end;
      readln(f,ligne);


      if Mot_cle('$noe',ligne)
        then
          begin
            readln(f,NNoeuds);
            for i:=1 to NNoeuds do readln(f,Noeuds[i]);
          end;
      if Mot_cle('$ele',ligne)
        then
          begin
            readln(f,Elements.Nombre);
            for i:=1 to Elements.Nombre do
              begin
                Elements.Table[i]:=TElement.Create;
                with Elements.Table[i] do
                  begin
                    readln(f,Origine,Extremite,Young,Iz);
                    if Longueur<1e-6
                      then
                        begin
                          writeln('La longueur de l''lment ',i,' est nulle');
                          result:=False;Close(f);Exit;
                        end;
                    if Iz<1e-12
                      then
                        begin
                          writeln('Le moment quadratique Iz de l''lment ',i,' est nulle');
                          result:=False;Close(f);Exit;
                        end;
                  end;
              end;
          end;
      if Mot_cle('$lia',ligne)
        then
          begin
            readln(f,Liaisons.Nombre);
            for i:=1 to Liaisons.Nombre do
              begin
                Liaisons.Table[i]:=TLiaison.Create;
                with Liaisons.Table[i] do readln(f,Type_liaison,Noeud);
              end;
          end;
      if Mot_cle('$cha',ligne)
        then
          begin
            readln(f,Charges.Nombre);
            for i:=1 to Charges.Nombre do
              begin
                Charges.Table[i]:=TCharge.Create;
                with Charges.Table[i] do
                  begin
                    read(f,Type_charge,Entite,Fy);
                    if Type_charge=id_force_nodale then read(f,Mz);
                    readln(f);
                  end;
              end;
          end;

    until Mot_cle('$fin',ligne) or Eof(f);

    close(f);

  end;

procedure TStructure.Localisation_ddl_noeud(i:Integer;var ddl_uy,ddl_rotz:Integer);
  begin
    ddl_uy:=2*i-1;
    ddl_rotz:=ddl_uy+1;
  end;

procedure TStructure.Localisation_ddl_element(i:Integer;
                            var locddl:TLocalisation_ddl_elementaires);
  var element:TElement;
  begin
    element:=Elements.Table[i];
    Localisation_ddl_noeud(element.Origine,locddl[1],locddl[2]);
    Localisation_ddl_noeud(element.Extremite,locddl[3],locddl[4]);
  end;

function TStructure.Assembler_la_matrice_de_rigidite:Boolean;

  var ielement,ice,ile,icolonne,iligne:Integer;
      element:TElement;
      ke:TMatrice_elementaire;
      locddl:TLocalisation_ddl_elementaires;

  begin

    writeln(fres,'Assemblage de la matrice de rigidit');

    result:=False;

    NDdl:=2*NNoeuds;
    for iligne:=1 to NDdl do
      for icolonne:=1 to NDdl do Matkg[iligne,icolonne]:=0;

    for ielement:=1 to Elements.Nombre do
      begin
        element:=Elements.Table[ielement];
        if not element.Rigidite(ke) then exit;
        Localisation_ddl_element(ielement,locddl);
        for ile:=1 to NDDLE do
          for ice:=1 to NDDLE do
            begin
              iligne:=locddl[ile];icolonne:=locddl[ice];
              Matkg[iligne,icolonne]:=Matkg[iligne,icolonne]+ke[ile,ice]
            end;
       end;

    result:=True;

  end;

procedure TStructure.Assembler_le_vecteur_force;

   var ielement,icharge,k,ddl_uy,ddl_rotz:Integer;
       locddl:TLocalisation_ddl_elementaires;
       feq:TVecteur_elementaire;
       long:Double;
       element:TElement;
       charge:TCharge;

   begin

     writeln(fres,'Assemblage du vecteur force global');

     for k:=1 to NDdl do Vectg[k]:=0;
     for ielement:=1 to Elements.Nombre do
       with Elements.Table[ielement] do
         for k:=1 to 4 do Fe[k]:=0;
     for icharge:=1 to Charges.Nombre do
       begin
         charge:=Charges.Table[icharge];
         with charge do
           if Type_charge=id_force_nodale
             then
               begin
                 Localisation_ddl_noeud(Entite,ddl_uy,ddl_rotz);
                 Vectg[ddl_uy]:=Vectg[ddl_uy]+Fy;
                 Vectg[ddl_rotz]:=Vectg[ddl_rotz]+Mz;
               end
             else
               begin
                 Localisation_ddl_element(Entite,locddl);
                 element:=Elements.Table[Entite];
                 long:=element.Longueur;
                 feq[1]:=-Fy*long/2;feq[2]:=-Fy*sqr(long)/12;
                 feq[3]:=feq[1];feq[4]:=-feq[2];
                 for k:=1 to 4 do Vectg[locddl[k]]:=Vectg[locddl[k]]-feq[k];
                 with element do for k:=1 to 4 do Fe[k]:=Fe[k]+feq[k];
               end;
       end;

   end;

procedure TStructure.Limites;  // mise en place des Liaisons

   var i,ddl_uy,ddl_rotz:Integer;

   procedure modifier_ligne_colonne(k:Integer);
      var j:Integer;
      begin
        for j:=1 to NDdl do Matkg[j,k]:=0;
        for j:=1 to NDdl do Matkg[k,j]:=0;
        Matkg[k,k]:=1;Vectg[k]:=0;
      end;

   begin

     writeln(fres,'Mise en place des Liaisons');

     for i:=1 to Liaisons.Nombre do
       with Liaisons.Table[i] do
         begin
           Localisation_ddl_noeud(Noeud,ddl_uy,ddl_rotz);
           case Type_liaison of
             ID_DY_NUL:Modifier_ligne_colonne(ddl_uy);
             ID_ROTZ_NUL:Modifier_ligne_colonne(ddl_rotz);
             ID_ENCASTREMENT:
               begin
                 Modifier_ligne_colonne(ddl_uy);
                 Modifier_ligne_colonne(ddl_rotz);
               end;
           end;
         end;

   end;

function TStructure.Calculer_les_deplacements:Boolean;

   // rsolution du systme linaire symtrique par la mthode de Gauss

   const eps=1e-10;
   
   var   iligne,il,icol:Integer;
         c:Double;

   procedure Erreur;
     begin
       writeln(fres,'Le systme est mal conditionn : pivot ngatif ou nul');
       result:=False;
     end;

   begin

     writeln(fres,'Calcul des dplacements inconnus');

     result:=True;

     // triangulation

     for iligne:=1 to NDdl-1 do
       begin
         if Matkg[iligne,iligne]<eps then begin Erreur;Exit;end;
         for il:=iligne+1 to NDdl do
           begin
             c:=Matkg[iligne,il]/Matkg[iligne,iligne];
             Vectg[il]:=Vectg[il]-c*Vectg[iligne];
             for icol:=il to NDdl do
               Matkg[il,icol]:=Matkg[il,icol]-c*Matkg[iligne,icol];
           end;
       end;

     // rsolution du systme triangulaire suprieur

     if Matkg[NDdl,NDdl]<eps then begin Erreur;Exit;end;
     Vectg[NDdl]:=Vectg[NDdl]/Matkg[NDdl,NDdl];
     for iligne:=NDdl-1 downto 1 do
       begin
         c:=0;
         for icol:=iligne+1 to NDdl do
           c:=c+Matkg[iligne,icol]*Vectg[icol];
         Vectg[iligne]:=(Vectg[iligne]-c)/Matkg[iligne,iligne];
       end;

   end;

procedure TStructure.Efforts_elementaires;

   var i,k,j:Integer;
       depl:TVecteur_elementaire;
       locddl:TLocalisation_ddl_elementaires;
       ke:TMatrice_elementaire;
       element:TElement;

   begin

     writeln(fres,'Calcul des forces lmentaires');

     for i:=1 to Elements.Nombre do
       begin
         Localisation_ddl_element(i,locddl);
         for k:=1 to NDDLE do depl[k]:=Vectg[locddl[k]];
         element:=Elements.Table[i];
         with element do
           begin
             Rigidite(ke);
             for k:=1 to NDDLE do
               for j:=1 to 4 do Fe[k]:=Fe[k]+ke[k,j]*depl[j];
           end;
       end;

    end;

procedure TStructure.Actions_de_liaison;

   var iliaison,ielement:Integer;
       liaison:TLiaison;
       element:TElement;
       fy,mz:Double;

   begin
     writeln(fres,'Calcul des actions de liaison');
     for iliaison:=1 to Liaisons.Nombre do
       begin
         liaison:=Liaisons.Table[iliaison];
         fy:=0;mz:=0;
         for ielement:=1 to Elements.Nombre do
           begin
             element:=Elements.Table[ielement];
             if element.Origine=liaison.Noeud
               then
                 begin
                   fy:=fy+element.Fe[1];
                   mz:=Mz+element.Fe[2];
                 end;
             if element.Extremite=liaison.Noeud
               then
                 begin
                   fy:=fy+element.Fe[3];
                   mz:=Mz+element.Fe[4];
                 end;
           end;
         liaison.Ry:=fy;
         liaison.Mz:=mz;
       end;
   end;

procedure TStructure.Editions;

   var i,k:Integer;
       g:TGraphe_elementaire;

   begin

     writeln(fres);

     writeln(fres,'Calcul d''une poutre droite');writeln(fres);
     writeln(fres,'Nom du fichier de donnes : ',nom_du_projet,'.txt');
     writeln(fres);

     writeln(fres,'Noeuds :');writeln(fres);
     for i:=1 to NNoeuds do writeln(fres,i:2,'  x = ',Noeuds[i]:10:3,' m');
     writeln(fres);

     writeln(fres,'lments :');writeln(fres);
     for i:=1 to Elements.Nombre do
       with Elements.Table[i] do
         begin
           write(fres,i:2,' ( ',Origine:2,Extremite:3,' )');
           writeln(fres,'  E = ',Young*1e-6:5:0,'E6 Pa   Iz = ',Iz*1e8:5:2,' cm4');
         end;
     writeln(fres);

     writeln(fres,'Liaisons :');writeln(fres);
     for i:=1 to Liaisons.Nombre do
       with Liaisons.table[i] do
         begin
           write(fres,'noeud ',Noeud:2);
           case type_liaison of
             ID_DY_NUL:writeln(fres,'   flche = 0');
             ID_ROTZ_NUL:writeln(fres,'   pente = 0');
             ID_ENCASTREMENT:writeln(fres,'   encastrement');
           end;
         end;
     writeln(fres);

     writeln(fres,'Charges :');writeln(fres);
     for i:=1 to Charges.Nombre do
       with Charges.Table[i] do
         if Type_charge=id_force_nodale
           then
             begin
               write(fres,'noeud ',Entite:2);
               if Abs(Fy)>1e-6 then write(fres,'  Fy = ',Fy:7:0,' N  ');
               if Abs(Mz)>1e-6 then write(fres,'  Mz = ',Mz:7:0,' N.m');
               writeln(fres);
             end
           else writeln(fres,'lment ',Entite:2,' : py = ',Fy:10:0,' N/m');
     writeln(fres);

     writeln(fres,'Dplacements nodaux ( m , rad ):');writeln(fres);
     writeln(fres,'Noeud','Flche':8,'Pente':10);writeln(fres);
     for i:=1 to NNoeuds do
       writeln(fres,i:3,' ',Vectg[2*i-1]:10:6,' ',Vectg[2*i]:10:6);
     writeln(fres);

     writeln(fres,'Efforts dans les lments ( N , N.m ) :');writeln(fres);
     for i:=1 to Elements.Nombre do
       with Elements.Table[i] do
         begin
           write(fres,i:4);
           writeln(fres,Origine:4,'    Ty = ',-Fe[1]:5:0,'  Mfz = ',-Fe[2]:5:0);
           writeln(fres,Extremite:8,'    Ty = ',Fe[3]:5:0,'  Mfz = ',Fe[4]:5:0);
       end;

     writeln(fres);
     writeln(fres,'Actions de liaison ( N , N.m ) :');writeln(fres);
     for i:=1 to Liaisons.Nombre do
       with Liaisons.table[i] do
         begin
           write(fres,'noeud ',Noeud,' : ');
           case Type_liaison of
             ID_DY_NUL:
               writeln(fres,'Ry = ',Ry:5:0);
             ID_ROTZ_NUL:
               writeln(fres,'Mz = ',Mz:5:0);
             ID_ENCASTREMENT:
               writeln(fres,'Ry = ',Ry:5:0,'  Mz = ',Mz:5:0);
           end;
         end;

     writeln(fres);
     writeln(fres,'Graphes lmentaires :');
     for i:=1 to Elements.Nombre do
       begin
         writeln(fres);writeln(fres,'lment ',i,' :');writeln(fres);
         writeln(fres,'x(m)':6,'Flche(m)':15,'Pente(rad)':13,'Ty(N)':9,'Mfz(N.m)':12);writeln(fres);
         Elements.Table[i].Graphe(g);
         for k:=0 to g.NPoints do
           begin
             write(fres,g.x[k]:7:3,g.depl_y[k]:12:6,g.rot_z[k]:13:5);
             writeln(fres,g.ty[k]:12:2,g.mfz[k]:10:2);
           end;
       end;
     writeln(fres);


   end;

procedure TStructure.Calculer;
  begin
     if not Lire_les_donnees then Exit;
     if not Assembler_la_matrice_de_rigidite then Exit;
     Assembler_le_vecteur_force;
     Limites;
     if not Calculer_les_deplacements then Exit;
     Efforts_elementaires;
     Actions_de_liaison;
     Editions;
  end;

//-------   programme principal  -----------

   begin

     writeln;
     writeln('  FLEXION SIMPLE D''UNE POUTRE DROITE');
     writeln('  PAR LA METHODE DES ELEMENTS FINIS');
     writeln('  (Modle de Bernoulli)');
     writeln;
     writeln('  Yves DEBARD - IUT du MANS - mai 2009');
     writeln;

     Structure:=TStructure.Create;
     Structure.Calculer;
     Structure.Free;

   end.
